/*****************************************************************************
 *
 *   Copyright(C) 2011, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "lwip/opt.h"
#include "lwip/def.h"
#include "fs.h"
#include "ff.h"
#include <string.h>

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

/* Define the number of open files that we can support. */
#ifndef LWIP_MAX_OPEN_FILES
#define LWIP_MAX_OPEN_FILES     10
#endif

/* Define the file system memory allocation structure. */
struct fs_table {
  struct fs_file file;
  u8_t inuse;
};

/* Allocate file system memory */
struct fs_table fs_memory[LWIP_MAX_OPEN_FILES];



/******************************************************************************
 *
 * Description:
 *    Allocate a file instance
 *
 *****************************************************************************/
static struct fs_file *
fs_malloc(void)
{
  int i;
  for(i = 0; i < LWIP_MAX_OPEN_FILES; i++) {
    if(fs_memory[i].inuse == 0) {
      fs_memory[i].inuse = 1;
      fs_memory[i].file.read = 0;
      return(&fs_memory[i].file);
    }
  }
  return(NULL);
}

/******************************************************************************
 *
 * Description:
 *    Free a file instance
 *
 *****************************************************************************/
static void
fs_free(struct fs_file* file)
{
  int i;
  for(i = 0; i < LWIP_MAX_OPEN_FILES; i++) {
    if(&fs_memory[i].file == file) {
      fs_memory[i].inuse = 0;
      break;
    }
  }
  return;
}

/******************************************************************************
 *
 * Description:
 *   Open a file
 *
 *****************************************************************************/
struct fs_file *
fs_open(const char *name)
{
  FRESULT result;
  struct fs_file *file;


  file = fs_malloc();
  if(file == NULL) {
    return NULL;
  }

  result = f_open(&file->f, name, FA_READ);
  if (result == FR_OK) {
    return file;
  }

  fs_free(file);
  return NULL;

}

/******************************************************************************
 *
 * Description:
 *    Close a file
 *
 *****************************************************************************/
void
fs_close(struct fs_file *file)
{
  if (file == NULL) {
    return;
  }

  f_close(&file->f);
  fs_free(file);
}

/******************************************************************************
 *
 * Description:
 *    Read data from a file
 *
 *****************************************************************************/
int
fs_read(struct fs_file *file, char *buffer, int count)
{

  FRESULT result;
  UINT numRead = 0;
  result = f_read (&file->f, buffer, count, &numRead);
  if (result != FR_OK) {
    return -1;
  }

  file->read += numRead;
  return numRead;
}

/******************************************************************************
 *
 * Description:
 *    Check number of unread bytes from an open file
 *
 *****************************************************************************/
int fs_bytes_left(struct fs_file *file)
{
  return (file->f.fsize - file->read);
}
